# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""


#: Holds the solarized style
SOLARIZED = {
    'base03': '#002b36',
    'base02': '#073642',
    'base01': '#586e75',
    'base00': '#657b83',
    'base0': '#839496',
    'base1': '#93a1a1',
    'base2': '#eee8d5',
    'base3': '#fdf6e3',
    'yellow': '#b58900',
    'orange': '#cb4b16',
    'red': '#dc322f',
    'magenta': '#d33682',
    'violet': '#6c71c4',
    'blue': '#268bd2',
    'cyan': '#2aa198',
    'green': '#859900'
}

#: Holds the monokai style
MONOKAI = {
    'ghostWhite': '#F8F8F0',
    'lightGhostWhite': '#F8F8F2',
    'lightGray': '#CCCCCC',
    'gray': '#888888',
    'brownGray': '#49483E',
    'darkGray': '#282828',
    'yellow': '#E6DB74',
    'blue': '#66D9EF',
    'magenta': '#F92672',
    'purple': '#AE81FF',
    'brown': '#75715E',
    'orange': '#FD971F',
    'lightOrange': '#FFD569',
    'green': '#A6E22E',
    'seaGreen': '#529B2F'
}



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import os
import sys
import copy
import types
import platform
from contextlib import contextmanager

from .core import Colorful
from . import terminal

#: Holds the current version
__version__ = '0.4.4'

# if we are on Windows we have to init colorama
if platform.system() == 'Windows':
    from colorama import init
    init()


class ColorfulModule(types.ModuleType):
    """
    Module Type object for dynamic attribute access on
    this module.
    """
    def __init__(self, colorful, *args):
        super(ColorfulModule, self).__init__(*args)
        self.colorful = colorful

    @contextmanager
    def with_setup(self, colormode=None, colorpalette=None, extend_colors=False):
        """
        Return a new Colorful object with the given color config.
        """
        colorful = Colorful(
            colormode=self.colorful.colormode,
            colorpalette=copy.copy(self.colorful.colorpalette)
        )

        colorful.setup(
            colormode=colormode, colorpalette=colorpalette, extend_colors=extend_colors
        )
        yield colorful

    @contextmanager
    def with_8_ansi_colors(self):
        yield Colorful(
            colormode=terminal.ANSI_8_COLORS,
            colorpalette=copy.copy(self.colorful.colorpalette)
        )

    @contextmanager
    def with_16_ansi_colors(self):
        yield Colorful(
            colormode=terminal.ANSI_16_COLORS,
            colorpalette=copy.copy(self.colorful.colorpalette)
        )

    @contextmanager
    def with_256_ansi_colors(self):
        yield Colorful(
            colormode=terminal.ANSI_256_COLORS,
            colorpalette=copy.copy(self.colorful.colorpalette)
        )

    @contextmanager
    def with_true_colors(self):
        yield Colorful(
            colormode=terminal.TRUE_COLORS,
            colorpalette=copy.copy(self.colorful.colorpalette)
        )

    @contextmanager
    def with_palette(self, colorpalette):
        yield Colorful(
            colormode=self.colorful.colormode,
            colorpalette=colorpalette
        )

    @contextmanager
    def with_updated_palette(self, colorpalette):
        colorful = Colorful(
            colormode=self.colorful.colormode,
            colorpalette=copy.copy(self.colorful.colorpalette),
        )
        colorful.update_palette(colorpalette)
        yield colorful

    @contextmanager
    def with_style(self, style_name):
        colorful = Colorful(
            colormode=self.colorful.colormode,
            colorpalette={},
        )
        colorful.use_style(style_name)
        yield colorful

    def __getattr__(self, name):
        """
        Dynamically get methods from Colorful object.
        """
        # check if original module had the requested attribute
        # if yes, we have to return it since we don't want to
        # break the module functionality
        orig_module = __name__ + '_orig'
        if orig_module in sys.modules:
            try:
                return getattr(sys.modules[orig_module], name)
            except AttributeError:
                pass  # fallback to colorful functionality

        return getattr(self.colorful, name)


if os.environ.get('COLORFUL_NO_MODULE_OVERWRITE', '0') != '1':
    # Unfortunately, prior to Python 3.4 the replaced
    # modules gets garbage collected which will
    # destroy all currently imported modules.
    # To prevent this, is to store a backup of the original
    # modules in sys.modules.
    sys.modules[__name__ + '_orig'] = sys.modules[__name__]
    # expose the Colorful class as module
    sys.modules[__name__] = ColorfulModule(Colorful(), __name__)



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

from __future__ import print_function
from __future__ import unicode_literals

import os
import sys

from . import ansi
from . import rgb
from . import styles
from . import terminal
from .utils import PY2, DEFAULT_ENCODING, UNICODE

#: Holds the name of the env variable which is
#  used as path to the default rgb.txt file
DEFAULT_RGB_TXT_PATH = os.environ.get(
    'COLORFUL_DEFAULT_COLOR_PALETTE',
    os.path.join(os.path.dirname(__file__), 'data', 'rgb.txt'))

#: Holds the color names mapped to RGB channels
COLOR_PALETTE = rgb.parse_rgb_txt_file(path=DEFAULT_RGB_TXT_PATH)


class ColorfulError(Exception):
    """
    Exception which is raised for Colorful specific
    usage errors.
    """
    pass


def translate_rgb_to_ansi_code(red, green, blue, offset, colormode):
    """
    Translate the given RGB color into the appropriate ANSI escape code
    for the given color mode.
    The offset is used for the base color which is used.

    The ``colormode`` has to be one of:
        * 0: no colors / disabled
        * 8: use ANSI 8 colors
        * 16: use ANSI 16 colors (same as 8 but with brightness)
        * 256: use ANSI 256 colors
        * 0xFFFFFF / 16777215: use 16 Million true colors

    :param int red: the red channel value
    :param int green: the green channel value
    :param int blue: the blue channel value
    :param int offset: the offset to use for the base color
    :param int colormode: the color mode to use. See explanation above
    """
    if colormode == terminal.NO_COLORS:  # colors are disabled, thus return empty string
        return '', ''

    if colormode == terminal.ANSI_8_COLORS or colormode == terminal.ANSI_16_COLORS:
        color_code = ansi.rgb_to_ansi16(red, green, blue)
        start_code = ansi.ANSI_ESCAPE_CODE.format(
            code=color_code + offset - ansi.FOREGROUND_COLOR_OFFSET)
        end_code = ansi.ANSI_ESCAPE_CODE.format(code=offset + ansi.COLOR_CLOSE_OFFSET)
        return start_code, end_code

    if colormode == terminal.ANSI_256_COLORS:
        color_code = ansi.rgb_to_ansi256(red, green, blue)
        start_code = ansi.ANSI_ESCAPE_CODE.format(code='{base};5;{code}'.format(
            base=8 + offset, code=color_code))
        end_code = ansi.ANSI_ESCAPE_CODE.format(code=offset + ansi.COLOR_CLOSE_OFFSET)
        return start_code, end_code

    if colormode == terminal.TRUE_COLORS:
        start_code = ansi.ANSI_ESCAPE_CODE.format(code='{base};2;{red};{green};{blue}'.format(
            base=8 + offset, red=red, green=green, blue=blue))
        end_code = ansi.ANSI_ESCAPE_CODE.format(code=offset + ansi.COLOR_CLOSE_OFFSET)
        return start_code, end_code

    raise ColorfulError('invalid color mode "{0}"'.format(colormode))


def translate_colorname_to_ansi_code(colorname, offset, colormode, colorpalette):
    """
    Translate the given color name to a valid
    ANSI escape code.

    :parma str colorname: the name of the color to resolve
    :parma str offset: the offset for the color code
    :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``
    :parma dict colorpalette: the color palette to use for the color name mapping

    :returns str: the color as ANSI escape code

    :raises ColorfulError: if the given color name is invalid
    """
    try:
        red, green, blue = colorpalette[colorname]
    except KeyError:
        raise ColorfulError('the color "{0}" is unknown. Use a color in your color palette (by default: X11 rgb.txt)'.format(  # noqa
            colorname))
    else:
        return translate_rgb_to_ansi_code(red, green, blue, offset, colormode)


def resolve_modifier_to_ansi_code(modifiername, colormode):
    """
    Resolve the given modifier name to a valid
    ANSI escape code.

    :param str modifiername: the name of the modifier to resolve
    :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``

    :returns str: the ANSI escape code for the modifier

    :raises ColorfulError: if the given modifier name is invalid
    """
    if colormode == terminal.NO_COLORS:  # return empty string if colors are disabled
        return '', ''

    try:
        start_code, end_code = ansi.MODIFIERS[modifiername]
    except KeyError:
        raise ColorfulError('the modifier "{0}" is unknown. Use one of: {1}'.format(
            modifiername, ansi.MODIFIERS.keys()))
    else:
        return ansi.ANSI_ESCAPE_CODE.format(
            code=start_code), ansi.ANSI_ESCAPE_CODE.format(
                code=end_code)


def translate_style(style, colormode, colorpalette):
    """
    Translate the given style to an ANSI escape code
    sequence.

    ``style`` examples are:

    * green
    * bold
    * red_on_black
    * bold_green
    * italic_yellow_on_cyan

    :param str style: the style to translate
    :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``
    :parma dict colorpalette: the color palette to use for the color name mapping
    """
    style_parts = iter(style.split('_'))

    ansi_start_sequence = []
    ansi_end_sequence = []

    try:
        # consume all modifiers
        part = None
        for mod_part in style_parts:
            part = mod_part
            if part not in ansi.MODIFIERS:
                break  # all modifiers have been consumed

            mod_start_code, mod_end_code = resolve_modifier_to_ansi_code(part, colormode)
            ansi_start_sequence.append(mod_start_code)
            ansi_end_sequence.append(mod_end_code)
        else:  # we've consumed all parts, thus we can exit
            raise StopIteration()

        # next part has to be a foreground color or the 'on' keyword
        # which means we have to consume background colors
        if part != 'on':
            ansi_start_code, ansi_end_code = translate_colorname_to_ansi_code(
                part, ansi.FOREGROUND_COLOR_OFFSET, colormode, colorpalette)
            ansi_start_sequence.append(ansi_start_code)
            ansi_end_sequence.append(ansi_end_code)
            # consume the required 'on' keyword after the foreground color
            next(style_parts)

        # next part has to be the background color
        part = next(style_parts)
        ansi_start_code, ansi_end_code = translate_colorname_to_ansi_code(
            part, ansi.BACKGROUND_COLOR_OFFSET, colormode, colorpalette)
        ansi_start_sequence.append(ansi_start_code)
        ansi_end_sequence.append(ansi_end_code)
    except StopIteration:  # we've consumed all parts of the styling string
        pass

    # construct and return ANSI escape code sequence
    return ''.join(ansi_start_sequence), ''.join(ansi_end_sequence)


def style_string(string, ansi_style, colormode, nested=False):
    """
    Style the given string according to the given
    ANSI style string.

    :param str string: the string to style
    :param tuple ansi_style: the styling string returned by ``translate_style``
    :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``

    :returns: a string containing proper ANSI sequence
    """
    ansi_start_code, ansi_end_code = ansi_style

    # replace nest placeholders with the current begin style
    if PY2:
        if isinstance(string, str):
            string = string.decode(DEFAULT_ENCODING)
    string = UNICODE(string).replace(ansi.NEST_PLACEHOLDER, ansi_start_code)

    return '{start_code}{string}{end_code}{nest_ph}'.format(
            start_code=ansi_start_code,
            string=string,
            end_code=ansi_end_code,
            nest_ph=ansi.NEST_PLACEHOLDER if nested else '')


class ColorfulString(object):
    """
    Represents a colored string
    """
    def __init__(self, orig_string, styled_string):
        self.orig_string = orig_string
        self.styled_string = styled_string

    if PY2:
        def __unicode__(self):
            return self.styled_string

        def __str__(self):
            return self.styled_string.encode(DEFAULT_ENCODING)
    else:
        def __str__(self):
            return self.styled_string

    def __len__(self):
        return len(self.orig_string)

    def __iter__(self):
        return iter(self.styled_string)

    def __add__(self, other):
        if isinstance(other, ColorfulString):
            return ColorfulString(
                self.orig_string + other.orig_string,
                self.styled_string + other.styled_string)

        return ColorfulString(
            self.orig_string + other,
            self.styled_string + other)

    def __iadd__(self, other):
        if isinstance(other, ColorfulString):
            self.orig_string += other.orig_string
            self.styled_string += other.styled_string
        else:
            self.orig_string += other
            self.styled_string += other

        return self

    def __radd__(self, other):
        if isinstance(other, ColorfulString):
            return ColorfulString(
                other.orig_string + self.orig_string,
                other.styled_string + self.styled_string)

        # we return handover the conversion to the
        # object on the left side
        return other + self.styled_string

    def __mul__(self, other):
        return ColorfulString(
            self.orig_string * other,
            self.styled_string * other)

    def __format__(self, format_spec):
        # append nested placeholder to styled string in order to continue the
        # previous styles. If the string already ends with the nest placeholder
        # we don't have to append it again.
        if self.styled_string.endswith(ansi.NEST_PLACEHOLDER):
            styled_string = self.styled_string
        else:
            styled_string = '{orig_str}{nest_ph}'.format(
                orig_str=self.styled_string, nest_ph=ansi.NEST_PLACEHOLDER)
        return styled_string.__format__(format_spec)

    def __getattr__(self, name):
        str_method = getattr(self.styled_string, name)
        return str_method


class Colorful(object):
    """
    Provides methods to style strings for terminal
    output.

    :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``
    """
    # re-expose the color modes from ``colorful.terminal``
    # on a package level.
    NO_COLORS = terminal.NO_COLORS
    ANSI_8_COLORS = terminal.ANSI_8_COLORS
    ANSI_16_COLORS = terminal.ANSI_16_COLORS
    ANSI_256_COLORS = terminal.ANSI_256_COLORS
    TRUE_COLORS = terminal.TRUE_COLORS

    # expose ANSI escape codes to close colors
    # this is especially useful when using ``str.format()``.
    close_fg_color = ansi.ANSI_ESCAPE_CODE.format(
        code=ansi.FOREGROUND_COLOR_OFFSET + ansi.COLOR_CLOSE_OFFSET)
    close_bg_color = ansi.ANSI_ESCAPE_CODE.format(
        code=ansi.BACKGROUND_COLOR_OFFSET + ansi.COLOR_CLOSE_OFFSET)

    # expose ANSI escape codes to close modifiers
    no_bold = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['bold'][1])
    no_dimmed = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['dimmed'][1])
    no_italic = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['italic'][1])
    no_underlined = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['underlined'][1])
    no_blinkslow = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['blinkslow'][1])
    no_blinkrapid = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['blinkrapid'][1])
    no_inversed = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['inversed'][1])
    no_concealed = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['concealed'][1])
    no_struckthrough = ansi.ANSI_ESCAPE_CODE.format(code=ansi.MODIFIERS['struckthrough'][1])

    def __init__(self, colormode=None, colorpalette=None):
        if colormode is None:  # try to auto-detect color mode
            colormode = terminal.detect_color_support(env=os.environ)

        if colorpalette is None:  # load default color palette
            colorpalette = COLOR_PALETTE
        elif isinstance(colorpalette, str):  # we assume it's a path to a X11 rgb.txt
            colorpalette = rgb.parse_rgb_txt_file(colorpalette)

        #: Holds the color mode to use for this Colorful object.
        self.colormode = colormode

        #: Holds the color palette to use for this Colorful object.
        self._colorpalette = None
        self.colorpalette = colorpalette

    @property
    def colorpalette(self):
        """
        Get the current used color palette
        """
        return self._colorpalette

    @colorpalette.setter
    def colorpalette(self, colorpalette):
        """
        Set the colorpalette which should be used
        """
        self._colorpalette = rgb.sanitize_color_palette(colorpalette)

    def setup(self, colormode=None, colorpalette=None, extend_colors=False):
        """
        Setup this colorful object by setting a ``colormode`` and
        the ``colorpalette`. The ``extend_colors`` flag is used
        to extend the currently active color palette instead of
        replacing it.

        :param int colormode: the color mode to use. See ``translate_rgb_to_ansi_code``
        :parma dict colorpalette: the colorpalette to use. This ``dict`` should map
                                  color names to it's corresponding RGB value
        :param bool extend_colors: extend the active color palette instead of replacing it
        """
        if colormode:
            self.colormode = colormode

        if colorpalette:
            if extend_colors:
                self.update_palette(colorpalette)
            else:
                self.colorpalette = colorpalette

    def disable(self):
        """
        Disable all colors and styles
        """
        self.colormode = terminal.NO_COLORS

    def use_8_ansi_colors(self):
        """
        Use 8 ANSI colors for this colorful object
        """
        self.colormode = terminal.ANSI_8_COLORS

    def use_16_ansi_colors(self):
        """
        Use 16 ANSI colors for this colorful object
        """
        self.colormode = terminal.ANSI_16_COLORS

    def use_256_ansi_colors(self):
        """
        Use 256 ANSI colors for this colorful object
        """
        self.colormode = terminal.ANSI_256_COLORS

    def use_true_colors(self):
        """
        Use true colors for this colorful object
        """
        self.colormode = terminal.TRUE_COLORS

    def use_palette(self, colorpalette):
        """
        Use the given color palette
        """
        self.colorpalette = colorpalette

    def update_palette(self, colorpalette):
        """
        Update the currently active color palette
        with the given color palette
        """
        self.colorpalette.update(rgb.sanitize_color_palette(colorpalette))

    def use_style(self, style_name):
        """
        Use a predefined style as color palette

        :param str style_name: the name of the style
        """
        try:
            style = getattr(styles, style_name.upper())
        except AttributeError:
            raise ColorfulError('the style "{0}" is undefined'.format(
                style_name))
        else:
            self.colorpalette = style

    def format(self, string, *args, **kwargs):
        """
        Format the given string with the given ``args`` and ``kwargs``.
        The string can contain references to ``c`` which is provided by
        this colorful object.

        :param str string: the string to format
        """
        return string.format(c=self, *args, **kwargs)

    def str(self, string):
        """
        Create a new ColorfulString instance of the given
        unstyled string.

        This method should be used to create a ColorfulString
        which is actually not styled yet but can safely be concatinated
        with other ColorfulStrings like:

        >>> s = colorful.str('Hello ')
        >>> s =+ colorful.black('World')
        >>> str(s)
        'Hello \033[30mWorld\033[39m'

        :param str string: the string to use for the ColorfulString
        """
        return ColorfulString(string, string)

    def print(self, *objects, **options):
        """
        Print the given objects to the given file stream.
        See https://docs.python.org/3/library/functions.html#print

        The only difference to the ``print()`` built-in is that
        ``Colorful.print()`` formats the string with ``c=self``.
        With that stylings are possible

        :param str sep: the seperater between the objects
        :param str end: the ending delimiter after all objects
        :param file: the file stream to write to
        :param bool flush: if the stream should be flushed
        """
        # NOTE: change signature to same as print() built-in function as
        #       soon as Python 2.7 is not supported anymore.
        #       If causes problems because of the keyword args after *args
        allowed_options = {'sep', 'end', 'file', 'flush'}
        given_options = set(options.keys())
        if not given_options.issubset(allowed_options):
            raise TypeError('Colorful.print() got unexpected keyword arguments: {0}'.format(
                ', '.join(given_options.difference(allowed_options))))

        sep = options.get('sep', ' ')
        end = options.get('end', '\n')
        file = options.get('file', sys.stdout)
        flush = options.get('flush', False)

        styled_objects = [self.format(o) for o in objects]
        print(*styled_objects, sep=sep, end=end, file=file)

        # NOTE: if Python 2.7 support is dropped we can directly forward the
        #       flush keyword argument to the print() function.
        if flush:
            file.flush()

    class ColorfulStyle(object):
        """
        Represents a colorful style
        """
        def __init__(self, style, colormode):
            self.style = style
            self.colormode = colormode

        def evaluate(self, string, nested=False):
            """
            Evaluate the style on the given string.

            :parma str string: the string to style
            :param bool nested: if the string is part of another styled string
                                (=> nested in another style)
            """
            return ColorfulString(
                string,
                style_string(string, self.style, self.colormode, nested))

        def __str__(self):
            return self.style[0]

        def __and__(self, other):
            new_style = (
                self.style[0] + other.style[0],
                self.style[1] + other.style[1]
            )
            return Colorful.ColorfulStyle(new_style, self.colormode)

        def __call__(self, string, nested=False):
            return self.evaluate(string, nested)

        def __or__(self, other):
            return self.evaluate(other)

    def __getattr__(self, name):
        # translate the given name into an ANSI escape code sequence
        style = translate_style(name, self.colormode, self.colorpalette)
        style_wrapper = self.ColorfulStyle(style, self.colormode)
        return style_wrapper



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import sys


# Valid color modes for colorful
NO_COLORS = 0
ANSI_8_COLORS = 8
ANSI_16_COLORS = 16
ANSI_256_COLORS = 256
TRUE_COLORS = 0xFFFFFF


def detect_color_support(env):  # noqa
    """
    Detect what color palettes are supported.
    It'll return a valid color mode to use
    with colorful.

    :param dict env: the environment dict like returned by ``os.envion``
    """
    if env.get('COLORFUL_DISABLE', '0') == '1':
        return NO_COLORS

    if env.get('COLORFUL_FORCE_8_COLORS', '0') == '1':
        return ANSI_8_COLORS

    if env.get('COLORFUL_FORCE_16_COLORS', '0') == '1':
        return ANSI_16_COLORS

    if env.get('COLORFUL_FORCE_256_COLORS', '0') == '1':
        return ANSI_256_COLORS

    if env.get('COLORFUL_FORCE_TRUE_COLORS', '0') == '1':
        return TRUE_COLORS

    # if we are not a tty
    if not sys.stdout.isatty():
        return NO_COLORS

    colorterm_env = env.get('COLORTERM')
    if colorterm_env:
        if colorterm_env in {'truecolor', '24bit'}:
            return TRUE_COLORS

        if colorterm_env in {'8bit'}:
            return ANSI_256_COLORS

    termprog_env = env.get('TERM_PROGRAM')
    if termprog_env:
        if termprog_env in {'iTerm.app', 'Hyper'}:
            return TRUE_COLORS

        if termprog_env in {'Apple_Terminal'}:
            return ANSI_256_COLORS

    term_env = env.get('TERM')
    if term_env:
        if term_env in {'screen-256', 'screen-256color', 'xterm-256', 'xterm-256color'}:
            return ANSI_256_COLORS

        if term_env in {'screen', 'xterm', 'vt100', 'color', 'ansi', 'cygwin', 'linux'}:
            return ANSI_16_COLORS

    if colorterm_env:
        # if there was no match with $TERM either but we
        # had one with $COLORTERM, we use it!
        return ANSI_16_COLORS

    return ANSI_8_COLORS



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import math

# For the ANSI escape code sequences please consult
# https://en.wikipedia.org/wiki/ANSI_escape_code

#: Holds the modifier names in the correct order
MODIFIERS = {
    'reset': (0, 0),
    'bold': (1, 22),
    'dimmed': (2, 22),
    'italic': (3, 23),
    'underlined': (4, 24),
    'blinkslow': (5, 25),
    'blinkrapid': (6, 25),
    'inversed': (7, 27),
    'concealed': (8, 28),
    'struckthrough': (9, 29)
}

#: Holds the offset for resetting the modifiers
MODIFIER_RESET_OFFSET = 21

#: Holds the start index for the fore- and background colors
FOREGROUND_COLOR_OFFSET = 30
BACKGROUND_COLOR_OFFSET = 40

#: Holds the offset to reset the color back to default.
#  this number has to be added to either ``FOREGROUND_COLOR_OFFSET`` or
#  ``BACKGROUND_COLOR_OFFSET`` to get an usable ANSI escape code.
COLOR_CLOSE_OFFSET = 9

#: Holds the Control Sequence Introducer
CSI = '\033['

#: Holds the base ANSI escape code
ANSI_ESCAPE_CODE = '{csi}{{code}}m'.format(csi=CSI)

#: Holds the placeholder for the nest indicators
NEST_PLACEHOLDER = ANSI_ESCAPE_CODE.format(code=26)


def round(value):
    x = math.floor(value)
    if (value - x) < 0.5:
        return int(x)
    return int(math.ceil(value))


def rgb_to_ansi256(r, g, b):
    """
    Convert RGB to ANSI 256 color
    """
    if r == g and g == b:
        if r < 8:
            return 16
        if r > 248:
            return 231

        return round(((r - 8) / 247.0) * 24) + 232

    ansi_r = 36 * round(r / 255.0 * 5.0)
    ansi_g = 6 * round(g / 255.0 * 5.0)
    ansi_b = round(b / 255.0 * 5.0)
    ansi = 16 + ansi_r + ansi_g + ansi_b
    return ansi


def rgb_to_ansi16(r, g, b, use_bright=False):
    """
    Convert RGB to ANSI 16 color
    """
    ansi_b = round(b / 255.0) << 2
    ansi_g = round(g / 255.0) << 1
    ansi_r = round(r / 255.0)
    ansi = (90 if use_bright else 30) + (ansi_b | ansi_g | ansi_r)

    return ansi



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

# NOTE: only used for Colorful.print()
from __future__ import print_function

import os

import pytest

# do not overwrite module
os.environ['COLORFUL_NO_MODULE_OVERWRITE'] = '1'

import colorful.core as core  # noqa
import colorful.terminal as terminal  # noqa
from colorful.utils import UNICODE  # noqa


@pytest.mark.parametrize('style_string,expected', [
    # modifiers
    ('bold', ('\033[1m', '\033[22m')),
    ('struckthrough', ('\033[9m', '\033[29m')),
    # foreground colors
    ('black', ('\033[30m', '\033[39m')),
    ('blue', ('\033[34m', '\033[39m')),
    ('white', ('\033[37m', '\033[39m')),
    # background colors
    ('on_black', ('\033[40m', '\033[49m')),
    ('on_blue', ('\033[44m', '\033[49m')),
    ('on_white', ('\033[47m', '\033[49m')),
    # modifiers with foreground colors
    ('bold_black', ('\033[1m\033[30m', '\033[22m\033[39m')),
    ('italic_blue', ('\033[3m\033[34m', '\033[23m\033[39m')),
    ('struckthrough_white', ('\033[9m\033[37m', '\033[29m\033[39m')),
    # modifiers with background colors
    ('bold_on_black', ('\033[1m\033[40m', '\033[22m\033[49m')),
    ('italic_on_blue', ('\033[3m\033[44m', '\033[23m\033[49m')),
    ('struckthrough_on_white', ('\033[9m\033[47m', '\033[29m\033[49m')),
    # modifiers with foreground and background colors
    ('bold_green_on_black', ('\033[1m\033[32m\033[40m', '\033[22m\033[39m\033[49m')),
    ('italic_cyan_on_blue', ('\033[3m\033[36m\033[44m', '\033[23m\033[39m\033[49m')),
    ('struckthrough_yellow_on_white', ('\033[9m\033[33m\033[47m', '\033[29m\033[39m\033[49m')),
    # multiple modifiers
    ('bold_italic', ('\033[1m\033[3m', '\033[22m\033[23m')),
    ('underlined_struckthrough', ('\033[4m\033[9m', '\033[24m\033[29m')),
    # multiple modifiers with foreground colors
    ('bold_italic_green', ('\033[1m\033[3m\033[32m', '\033[22m\033[23m\033[39m')),
    ('underlined_struckthrough_cyan', ('\033[4m\033[9m\033[36m', '\033[24m\033[29m\033[39m'))
])
def test_translate_style_8(style_string, expected):
    """
    Test translating style strings with 8 colors
    """
    assert core.translate_style(style_string, colormode=terminal.ANSI_8_COLORS,
                                colorpalette=core.COLOR_PALETTE) == expected


@pytest.mark.parametrize('style_string,expected', [
    # modifiers
    ('bold', ('\033[1m', '\033[22m')),
    ('struckthrough', ('\033[9m', '\033[29m')),
    # foreground colors
    ('black', ('\033[30m', '\033[39m')),
    ('blue', ('\033[34m', '\033[39m')),
    ('white', ('\033[37m', '\033[39m')),
    # background colors
    ('on_black', ('\033[40m', '\033[49m')),
    ('on_blue', ('\033[44m', '\033[49m')),
    ('on_white', ('\033[47m', '\033[49m')),
    # modifiers with foreground colors
    ('bold_black', ('\033[1m\033[30m', '\033[22m\033[39m')),
    ('italic_blue', ('\033[3m\033[34m', '\033[23m\033[39m')),
    ('struckthrough_white', ('\033[9m\033[37m', '\033[29m\033[39m')),
    # modifiers with background colors
    ('bold_on_black', ('\033[1m\033[40m', '\033[22m\033[49m')),
    ('italic_on_blue', ('\033[3m\033[44m', '\033[23m\033[49m')),
    ('struckthrough_on_white', ('\033[9m\033[47m', '\033[29m\033[49m')),
    # modifiers with foreground and background colors
    ('bold_green_on_black', ('\033[1m\033[32m\033[40m', '\033[22m\033[39m\033[49m')),
    ('italic_cyan_on_blue', ('\033[3m\033[36m\033[44m', '\033[23m\033[39m\033[49m')),
    ('struckthrough_yellow_on_white', ('\033[9m\033[33m\033[47m', '\033[29m\033[39m\033[49m')),
    # multiple modifiers
    ('bold_italic', ('\033[1m\033[3m', '\033[22m\033[23m')),
    ('underlined_struckthrough', ('\033[4m\033[9m', '\033[24m\033[29m')),
    # multiple modifiers with foreground colors
    ('bold_italic_green', ('\033[1m\033[3m\033[32m', '\033[22m\033[23m\033[39m')),
    ('underlined_struckthrough_cyan', ('\033[4m\033[9m\033[36m', '\033[24m\033[29m\033[39m'))
])
def test_translate_style_16(style_string, expected):
    """
    Test translating style strings with 16 colors
    """
    assert core.translate_style(style_string, colormode=terminal.ANSI_16_COLORS,
                                colorpalette=core.COLOR_PALETTE) == expected


@pytest.mark.parametrize('style_string,expected', [
    # modifiers
    ('bold', ('\033[1m', '\033[22m')),
    ('struckthrough', ('\033[9m', '\033[29m')),
    # foreground colors
    ('black', ('\033[38;5;16m', '\033[39m')),
    ('blue', ('\033[38;5;21m', '\033[39m')),
    ('white', ('\033[38;5;231m', '\033[39m')),
    # background colors
    ('on_black', ('\033[48;5;16m', '\033[49m')),
    ('on_blue', ('\033[48;5;21m', '\033[49m')),
    ('on_white', ('\033[48;5;231m', '\033[49m')),
    # modifiers with foreground colors
    ('bold_black', ('\033[1m\033[38;5;16m', '\033[22m\033[39m')),
    ('italic_blue', ('\033[3m\033[38;5;21m', '\033[23m\033[39m')),
    ('struckthrough_white', ('\033[9m\033[38;5;231m', '\033[29m\033[39m')),
    # modifiers with background colors
    ('bold_on_black', ('\033[1m\033[48;5;16m', '\033[22m\033[49m')),
    ('italic_on_blue', ('\033[3m\033[48;5;21m', '\033[23m\033[49m')),
    ('struckthrough_on_white', ('\033[9m\033[48;5;231m', '\033[29m\033[49m')),
    # modifiers with foreground and background colors
    ('bold_green_on_black', ('\033[1m\033[38;5;46m\033[48;5;16m', '\033[22m\033[39m\033[49m')),
    ('italic_cyan_on_blue', ('\033[3m\033[38;5;51m\033[48;5;21m', '\033[23m\033[39m\033[49m')),
    ('struckthrough_yellow_on_white', ('\033[9m\033[38;5;226m\033[48;5;231m', '\033[29m\033[39m\033[49m')),  # noqa
    # multiple modifiers
    ('bold_italic', ('\033[1m\033[3m', '\033[22m\033[23m')),
    ('underlined_struckthrough', ('\033[4m\033[9m', '\033[24m\033[29m')),
    # multiple modifiers with foreground colors
    ('bold_italic_green', ('\033[1m\033[3m\033[38;5;46m', '\033[22m\033[23m\033[39m')),
    ('underlined_struckthrough_cyan', ('\033[4m\033[9m\033[38;5;51m', '\033[24m\033[29m\033[39m'))
])
def test_translate_style_256(style_string, expected):
    """
    Test translating style strings with 256 colors
    """
    assert core.translate_style(style_string, colormode=terminal.ANSI_256_COLORS,
                                colorpalette=core.COLOR_PALETTE) == expected


@pytest.mark.parametrize('style_string,expected', [
    # modifiers
    ('bold', ('\033[1m', '\033[22m')),
    ('struckthrough', ('\033[9m', '\033[29m')),
    # foreground colors
    ('black', ('\033[38;2;0;0;0m', '\033[39m')),
    ('blue', ('\033[38;2;0;0;255m', '\033[39m')),
    ('white', ('\033[38;2;255;255;255m', '\033[39m')),
    # background colors
    ('on_black', ('\033[48;2;0;0;0m', '\033[49m')),
    ('on_blue', ('\033[48;2;0;0;255m', '\033[49m')),
    ('on_white', ('\033[48;2;255;255;255m', '\033[49m')),
    # modifiers with foreground colors
    ('bold_black', ('\033[1m\033[38;2;0;0;0m', '\033[22m\033[39m')),
    ('italic_blue', ('\033[3m\033[38;2;0;0;255m', '\033[23m\033[39m')),
    ('struckthrough_white', ('\033[9m\033[38;2;255;255;255m', '\033[29m\033[39m')),
    # modifiers with background colors
    ('bold_on_black', ('\033[1m\033[48;2;0;0;0m', '\033[22m\033[49m')),
    ('italic_on_blue', ('\033[3m\033[48;2;0;0;255m', '\033[23m\033[49m')),
    ('struckthrough_on_white', ('\033[9m\033[48;2;255;255;255m', '\033[29m\033[49m')),
    # modifiers with foreground and background colors
    ('bold_green_on_black', ('\033[1m\033[38;2;0;255;0m\033[48;2;0;0;0m', '\033[22m\033[39m\033[49m')),  # noqa
    ('italic_cyan_on_blue', ('\033[3m\033[38;2;0;255;255m\033[48;2;0;0;255m', '\033[23m\033[39m\033[49m')),  # noqa
    ('struckthrough_yellow_on_white', ('\033[9m\033[38;2;255;255;0m\033[48;2;255;255;255m', '\033[29m\033[39m\033[49m')),  # noqa
    # multiple modifiers
    ('bold_italic', ('\033[1m\033[3m', '\033[22m\033[23m')),
    ('underlined_struckthrough', ('\033[4m\033[9m', '\033[24m\033[29m')),
    # multiple modifiers with foreground colors
    ('bold_italic_green', ('\033[1m\033[3m\033[38;2;0;255;0m', '\033[22m\033[23m\033[39m')),
    ('underlined_struckthrough_cyan', ('\033[4m\033[9m\033[38;2;0;255;255m', '\033[24m\033[29m\033[39m'))  # noqa
])
def test_translate_style_true_colors(style_string, expected):
    """
    Test translating style strings with true colors
    """
    assert core.translate_style(style_string, colormode=terminal.TRUE_COLORS,
                                colorpalette=core.COLOR_PALETTE) == expected


@pytest.mark.parametrize('method_name,expected', [
    ('bold', '\033[1mNo, I am your father\033[22m'),
    ('struckthrough', '\033[9mNo, I am your father\033[29m'),
    # foreground colors
    ('black', '\033[30mNo, I am your father\033[39m'),
    ('blue', '\033[34mNo, I am your father\033[39m'),
    ('white', '\033[37mNo, I am your father\033[39m'),
    # background colors
    ('on_black', '\033[40mNo, I am your father\033[49m'),
    ('on_blue', '\033[44mNo, I am your father\033[49m'),
    ('on_white', '\033[47mNo, I am your father\033[49m'),
    # modifiers with foreground colors
    ('bold_black', '\033[1m\033[30mNo, I am your father\033[22m\033[39m'),
    ('italic_blue', '\033[3m\033[34mNo, I am your father\033[23m\033[39m'),
    ('struckthrough_white', '\033[9m\033[37mNo, I am your father\033[29m\033[39m'),
    # modifiers with background colors
    ('bold_on_black', '\033[1m\033[40mNo, I am your father\033[22m\033[49m'),
    ('italic_on_blue', '\033[3m\033[44mNo, I am your father\033[23m\033[49m'),
    ('struckthrough_on_white', '\033[9m\033[47mNo, I am your father\033[29m\033[49m'),
    # modifiers with foreground and background colors
    ('bold_green_on_black', '\033[1m\033[32m\033[40mNo, I am your father\033[22m\033[39m\033[49m'),
    ('italic_cyan_on_blue', '\033[3m\033[36m\033[44mNo, I am your father\033[23m\033[39m\033[49m'),
    ('struckthrough_yellow_on_white', '\033[9m\033[33m\033[47mNo, I am your father\033[29m\033[39m\033[49m'),  # noqa
    # multiple modifiers
    ('bold_italic', '\033[1m\033[3mNo, I am your father\033[22m\033[23m'),
    ('underlined_struckthrough', '\033[4m\033[9mNo, I am your father\033[24m\033[29m'),
    # multiple modifiers with foreground colors
    ('bold_italic_green', '\033[1m\033[3m\033[32mNo, I am your father\033[22m\033[23m\033[39m'),
    ('underlined_struckthrough_cyan', '\033[4m\033[9m\033[36mNo, I am your father\033[24m\033[29m\033[39m')  # noqa
])
def test_method_call_to_style_conversion(method_name, expected):
    """
    Test converting the method call to an actual style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    method = getattr(colorful, method_name)

    assert str(method('No, I am your father')) == expected


@pytest.mark.parametrize('method_name', [
    ('bold'),
    ('struckthrough'),
    # foreground colors
    ('black'),
    ('blue'),
    ('white'),
    # background colors
    ('on_black'),
    ('on_blue'),
    ('on_white'),
    # modifiers with foreground colors
    ('bold_black'),
    ('italic_blue'),
    ('struckthrough_white'),
    # modifiers with background colors
    ('bold_on_black'),
    ('italic_on_blue'),
    ('struckthrough_on_white'),
    # modifiers with foreground and background colors
    ('bold_green_on_black'),
    ('italic_cyan_on_blue'),
    ('struckthrough_yellow_on_white'),
    # multiple modifiers
    ('bold_italic'),
    ('underlined_struckthrough'),
    # multiple modifiers with foreground colors
    ('bold_italic_green'),
    ('underlined_struckthrough_cyan'),
])
def test_method_call_to_style_conversion_disabled_colors(method_name):
    """
    Test converting the method call to an actual style with disabled colors
    """
    colorful = core.Colorful(colormode=terminal.NO_COLORS)
    method = getattr(colorful, method_name)

    assert str(method('No, I am your father')) == 'No, I am your father'


@pytest.mark.parametrize('method_name,expected', [
    ('bold', '\033[1m'),
    ('struckthrough', '\033[9m'),
    # foreground colors
    ('black', '\033[30m'),
    ('blue', '\033[34m'),
    ('white', '\033[37m'),
    # background colors
    ('on_black', '\033[40m'),
    ('on_blue', '\033[44m'),
    ('on_white', '\033[47m'),
    # modifiers with foreground colors
    ('bold_black', '\033[1m\033[30m'),
    ('italic_blue', '\033[3m\033[34m'),
    ('struckthrough_white', '\033[9m\033[37m'),
    # modifiers with background colors
    ('bold_on_black', '\033[1m\033[40m'),
    ('italic_on_blue', '\033[3m\033[44m'),
    ('struckthrough_on_white', '\033[9m\033[47m'),
    # modifiers with foreground and background colors
    ('bold_green_on_black', '\033[1m\033[32m\033[40m'),
    ('italic_cyan_on_blue', '\033[3m\033[36m\033[44m'),
    ('struckthrough_yellow_on_white', '\033[9m\033[33m\033[47m'),
    # multiple modifiers
    ('bold_italic', '\033[1m\033[3m'),
    ('underlined_struckthrough', '\033[4m\033[9m'),
    # multiple modifiers with foreground colors
    ('bold_italic_green', '\033[1m\033[3m\033[32m'),
    ('underlined_struckthrough_cyan', '\033[4m\033[9m\033[36m')
])
def test_method_str_to_style_conversion(method_name, expected):
    """
    Test converting the method to an actual style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    method = getattr(colorful, method_name)

    assert str(method) == expected


@pytest.mark.parametrize('format_str, expected', [
    (
        '{c.bold}No, I am your father{c.reset}',
        '\033[1mNo, I am your father\033[0m'
    ), (
        '{c.struckthrough}No, I am your father{c.reset}',
        '\033[9mNo, I am your father\033[0m'
    ),
    # foreground colors
    (
        '{c.black}No, I am your father{c.reset}',
        '\033[30mNo, I am your father\033[0m'
    ), (
        '{c.blue}No, I am your father{c.reset}',
        '\033[34mNo, I am your father\033[0m'
    ), (
        '{c.white}No, I am your father{c.reset}',
        '\033[37mNo, I am your father\033[0m'
    ),
    # background colors
    (
        '{c.on_black}No, I am your father{c.reset}',
        '\033[40mNo, I am your father\033[0m'
    ), (
        '{c.on_blue}No, I am your father{c.reset}',
        '\033[44mNo, I am your father\033[0m'
    ), (
        '{c.on_white}No, I am your father{c.reset}',
        '\033[47mNo, I am your father\033[0m'
    ),
    # modifiers with foreground colors
    (
        '{c.bold_black}No, I am your father{c.reset}',
        '\033[1m\033[30mNo, I am your father\033[0m'
    ), (
        '{c.italic_blue}No, I am your father{c.reset}',
        '\033[3m\033[34mNo, I am your father\033[0m'
    ), (
        '{c.struckthrough_white}No, I am your father{c.reset}',
        '\033[9m\033[37mNo, I am your father\033[0m'
    ),
    # modifiers with background colors
    (
        '{c.bold_on_black}No, I am your father{c.reset}',
        '\033[1m\033[40mNo, I am your father\033[0m'
    ), (
        '{c.italic_on_blue}No, I am your father{c.reset}',
        '\033[3m\033[44mNo, I am your father\033[0m'
    ), (
        '{c.struckthrough_on_white}No, I am your father{c.reset}',
        '\033[9m\033[47mNo, I am your father\033[0m'
    ),
    # modifiers with foreground and background colors
    (
        '{c.bold_green_on_black}No, I am your father{c.reset}',
        '\033[1m\033[32m\033[40mNo, I am your father\033[0m'
    ), (
        '{c.italic_cyan_on_blue}No, I am your father{c.reset}',
        '\033[3m\033[36m\033[44mNo, I am your father\033[0m'
    ), (
        '{c.struckthrough_yellow_on_white}No, I am your father{c.reset}',
        '\033[9m\033[33m\033[47mNo, I am your father\033[0m'
    ),
    # multiple modifiers
    (
        '{c.bold_italic}No, I am your father{c.reset}',
        '\033[1m\033[3mNo, I am your father\033[0m'
    ), (
        '{c.underlined_struckthrough}No, I am your father{c.reset}',
        '\033[4m\033[9mNo, I am your father\033[0m'
    ),
    # multiple modifiers with foreground colors
    (
        '{c.bold_italic_green}No, I am your father{c.reset}',
        '\033[1m\033[3m\033[32mNo, I am your father\033[0m'
    ), (
        '{c.underlined_struckthrough_cyan}No, I am your father{c.reset}',
        '\033[4m\033[9m\033[36mNo, I am your father\033[0m'
    ),
    # color closing delimiters
    (
        '{c.black}No, I am your father{c.close_fg_color}',
        '\033[30mNo, I am your father\033[39m'
    ), (
        '{c.on_black}No, I am your father{c.close_bg_color}',
        '\033[40mNo, I am your father\033[49m'
    ),
    # modifier closing delimiters
    (
        '{c.bold}No, I am your father{c.no_bold}',
        '\033[1mNo, I am your father\033[22m'
    ), (
        '{c.struckthrough}No, I am your father{c.no_struckthrough}',
        '\033[9mNo, I am your father\033[29m'
    )
])
def test_method_in_format_to_style_conversion(format_str, expected):
    """
    Test converting the method in a format() call to an actual style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    assert format_str.format(c=colorful) == expected


def test_invalid_color_mode():
    """
    Test setting an invalid color mode
    """
    colorful = core.Colorful(colormode=42)

    with pytest.raises(core.ColorfulError) as exc:
        colorful.white('Invalid color mode')
    assert str(exc.value) == 'invalid color mode "42"'


def test_invalid_color_name():
    """
    Test invalid color name
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    expected = ('the color "invalidColor" is unknown. '
                'Use a color in your color palette (by default: X11 rgb.txt)')
    with pytest.raises(core.ColorfulError) as exc:
        colorful.bold_invalidColor('Invalid color name')
    assert str(exc.value) == expected


@pytest.mark.parametrize('env,expected', [
    ({'COLORFUL_DISABLE': '1'}, terminal.NO_COLORS),
    ({'COLORTERM': 'truecolor'}, terminal.TRUE_COLORS),
    ({'TERM_PROGRAM': 'iTerm.app'}, terminal.TRUE_COLORS),
    ({'TERM': 'xterm-256color'}, terminal.ANSI_256_COLORS),
    ({'TERM': 'screen'}, terminal.ANSI_16_COLORS),
    ({}, terminal.ANSI_8_COLORS),
])
def test_colorful_obj_color_auto_detection(env, expected):
    """
    Test that the colorful object is able to auto detect the supported colors
    """
    os_env_backup = os.environ.copy()
    os.environ = env
    colorful = core.Colorful(colormode=None)  # None to explicitly auto detect the colors
    assert colorful.colormode == expected

    # restore environ backup
    os.environ = os_env_backup


def test_reading_color_palette(tmpdir):
    """
    Test reading color palette from file
    """
    # create palette file
    palette_file = tmpdir.mkdir('palettes').join('my_rgb.txt')
    palette_file.write("""
0 0 0 myBlack
255 255 255      myWhite""")

    colorful = core.Colorful(colorpalette=str(palette_file))
    assert str(colorful.myBlack) == '\033[30m'
    assert str(colorful.myWhite) == '\033[37m'


def test_colorful_obj_setup():
    """
    Test setup of an existing colorful object
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    colorful.setup(
        colormode=terminal.TRUE_COLORS,
        colorpalette={'testColor': (0, 0, 0)},
        extend_colors=False
    )

    assert colorful.colormode == terminal.TRUE_COLORS
    assert str(colorful.testColor) == '\033[38;2;0;0;0m'

    with pytest.raises(core.ColorfulError) as exc:
        colorful.black('black does not exist anymore')

    expected = ('the color "black" is unknown. '
                'Use a color in your color palette (by default: X11 rgb.txt)')
    assert str(exc.value) == expected


def test_colorful_obj_setup_with_extending_palette():
    """
    Test setup of an existing colorful object and extend the color palette
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    colorful.setup(
        colormode=terminal.TRUE_COLORS,
        colorpalette={'testColor': (0, 0, 0)},
        extend_colors=True
    )

    assert colorful.colormode == terminal.TRUE_COLORS
    assert str(colorful.testColor) == '\033[38;2;0;0;0m'
    assert str(colorful.black) == '\033[38;2;0;0;0m'


@pytest.mark.parametrize('method_name, colorname, expected', [
    ('use_8_ansi_colors', 'black', '\033[30m'),
    ('use_16_ansi_colors', 'black', '\033[30m'),
    ('use_256_ansi_colors', 'black', '\033[38;5;16m'),
    ('use_true_colors', 'black', '\033[38;2;0;0;0m')
])
def test_set_color_mode_methods(method_name, colorname, expected):
    """
    Test changing the color mode for an existing colorful object
    """
    colorful = core.Colorful(colormode=terminal.NO_COLORS)

    # change color mode
    getattr(colorful, method_name)()

    assert str(getattr(colorful, colorname)) == expected


def test_change_color_palette():
    """
    Test changing the color palette for an existing colorful object
    """
    NEW_COLOR_PALETTE = {
        'black': (0, 0, 0)
    }
    colorful = core.Colorful(
        colormode=terminal.ANSI_8_COLORS, colorpalette={'defaultColor': (255, 255, 255)})

    # updating existing color palette
    colorful.update_palette(NEW_COLOR_PALETTE)
    # old and new colors should exist
    assert str(colorful.black) == '\033[30m'
    assert str(colorful.defaultColor) == '\033[37m'

    # set color palette and overwrite the old one
    colorful.use_palette(NEW_COLOR_PALETTE)
    assert str(colorful.black) == '\033[30m'

    with pytest.raises(core.ColorfulError) as exc:
        colorful.defaultColor('The defaultColor was overwritten by the new palette')

    expected = ('the color "defaultColor" is unknown. '
                'Use a color in your color palette (by default: X11 rgb.txt)')
    assert str(exc.value) == expected


def test_use_styles():
    """
    Test using a predefined style
    """
    colorful = core.Colorful(colormode=terminal.TRUE_COLORS)
    colorful.use_style('solarized')

    assert str(colorful.red) == '\033[38;2;220;50;47m'

    with pytest.raises(core.ColorfulError) as exc:
        colorful.lightCoral('The color lightCoral only exists in the X11 rgb.txt palette')

    assert str(exc.value).startswith('the color "lightCoral" is unknown.')


def test_use_unknown_style():
    """
    Test exception for unknown style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    with pytest.raises(core.ColorfulError) as exc:
        colorful.use_style('unknown')

    assert str(exc.value) == 'the style "unknown" is undefined'


def test_colorful_format():
    """
    Test the colorful.format method
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    expected = '\033[3m\033[31mNo, I am your father\033[23m\033[39m'
    assert colorful.format('{c.italic_red}{0}, I am your {who}{c.no_italic}{c.close_fg_color}',
                           'No', who='father') == expected


def test_colorful_direct_print(capsys):
    """
    Test the colorful.print method
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    expected = u'\033[3m\033[31mNo, I am your father\033[23m\033[39m\n'

    colorful.print('{c.italic_red}No, I am your father{c.no_italic}{c.close_fg_color}', flush=True)

    out, err = capsys.readouterr()
    assert repr(out) == repr(expected)
    assert err == ''


def test_colorful_print_wrong_argument():
    """
    Test calling the colorful.print method with wrong arguments
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    with pytest.raises(TypeError) as exc:
        colorful.print('Never printed because of argument error', foo=42)

    assert str(exc.value) == 'Colorful.print() got unexpected keyword arguments: foo'


@pytest.mark.parametrize('method_name, expected', [
    ('bold', 'No, I am your father'),
    ('struckthrough', 'No, I am your father'),
    # foreground colors
    ('black', 'No, I am your father'),
    ('blue', 'No, I am your father'),
    ('white', 'No, I am your father'),
    # background colors
    ('on_black', 'No, I am your father'),
    ('on_blue', 'No, I am your father'),
    ('on_white', 'No, I am your father'),
    # modifiers with foreground colors
    ('bold_black', 'No, I am your father'),
    ('italic_blue', 'No, I am your father'),
    ('struckthrough_white', 'No, I am your father'),
    # modifiers with background colors
    ('bold_on_black', 'No, I am your father'),
    ('italic_on_blue', 'No, I am your father'),
    ('struckthrough_on_white', 'No, I am your father'),
    # modifiers with foreground and background colors
    ('bold_green_on_black', 'No, I am your father'),
    ('italic_cyan_on_blue', 'No, I am your father'),
    ('struckthrough_yellow_on_white', 'No, I am your father'),
    # multiple modifiers
    ('bold_italic', 'No, I am your father'),
    ('underlined_struckthrough', 'No, I am your father'),
    # multiple modifiers with foreground colors
    ('bold_italic_green', 'No, I am your father'),
    ('underlined_struckthrough_cyan', 'No, I am your father')
])
def test_length_of_styled_string(method_name, expected):
    """
    Test the length of a styled string
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    method = getattr(colorful, method_name)

    assert len(method(expected)) == len(expected)


def test_nested_styled_string():
    """
    Test nested styled string
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    s = colorful.red('Hello ' + colorful.blue('awesome', nested=True) + ' world')
    assert str(s) == '\033[31mHello \033[34mawesome\033[39m\033[31m world\033[39m'

    s = colorful.red('Hello') + ' World'
    assert str(s) == '\033[31mHello\033[39m World'

    s = 'Hello' + colorful.red(' World')
    assert str(s) == 'Hello\033[31m World\033[39m'

    s = colorful.red('Hello') + colorful.blue(' World')
    assert str(s) == '\033[31mHello\033[39m\033[34m World\033[39m'

    s = colorful.red('Hello {0} world'.format(colorful.blue('awesome', nested=True)))
    assert str(s) == '\033[31mHello \033[34mawesome\033[39m\033[31m world\033[39m'


def test_nested_styled_string_with_format():
    """
    Test nested styled string with format()
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    s = colorful.red('Hello {0} world'.format(colorful.blue('awesome')))
    assert str(s) == '\033[31mHello \033[34mawesome\033[39m\033[31m world\033[39m'


def test_styling_object_which_implements_str_proto():
    """
    Test styling an object which implements the str protocol
    """
    class Dummy(object):
        def __str__(self):
            return 'I am a dummy object'

    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    assert str(colorful.black(Dummy())) == '\033[30mI am a dummy object\033[39m'


def test_str_behavior_for_colorfulstring():
    """
    Test that the ColorfulString object behaves like a str
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    s = colorful.black('Hello')

    # test adding to str
    assert str(s + ' World') == '\033[30mHello\033[39m World'

    # test being added to str
    assert 'World ' + s == 'World \033[30mHello\033[39m'

    # test being augment added to str
    string = 'World '
    string += s
    assert string == 'World \033[30mHello\033[39m'

    # test augment add a str
    s2 = colorful.red('Hello ')
    s2 += 'World'
    assert str(s2) == '\033[31mHello \033[39mWorld'

    # test ColorfulString to ColorfulString
    assert str(s + s) == '\033[30mHello\033[39m\033[30mHello\033[39m'

    # test ColorfulString augmented added to other ColorfulString
    s2 = colorful.red('World ')
    s2 += s
    assert str(s2) == '\033[31mWorld \033[39m\033[30mHello\033[39m'

    # test multipling ColorfulString
    assert str(s * 3) == '\033[30mHello\033[39m\033[30mHello\033[39m\033[30mHello\033[39m'

    # other str methods operate on the styled string
    assert s.replace('Hello', 'Adieu') == '\033[30mAdieu\033[39m'


def test_joining_colorfulstrings():
    """
    Test joining ColorfulStrings
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    parts = [colorful.red('Hello'), colorful.black('World')]

    # FIXME(TF): is there any other way??
    string = ' '.join(str(part) for part in parts)
    assert string == '\033[31mHello\033[39m \033[30mWorld\033[39m'


def test_creating_unstyled_colorfulstring():
    """
    Test creating an unstyled ColorfulString
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    s = colorful.str('Hello ')
    s += colorful.black('World')

    assert s.orig_string == 'Hello World'
    assert s.styled_string == str(s) == 'Hello \033[30mWorld\033[39m'


def test_unicode_support():
    """
    Test unicode support
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    s = u'🐧🎉🐧'
    styled_s = colorful.black(s)

    # test basic unicode support
    assert UNICODE(styled_s) == u'\033[30m🐧🎉🐧\033[39m'


def test_combining_styles():
    """
    Test combining styles
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    style = colorful.bold & colorful.red & colorful.on_black

    assert str(style) == '\033[1m\033[31m\033[40m'


def test_piping_str_into_style():
    """
    Test piping a string into a style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    string = colorful.bold_red | 'No, I am your father'

    assert str(string) == '\033[1m\033[31mNo, I am your father\033[22m\033[39m'


def test_piping_styled_str_into_style():
    """
    Test piping a string into a style
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    string = colorful.bold_red | colorful.on_black('No, I am your father')

    assert str(string) == '\033[1m\033[31m\033[40mNo, I am your father\033[49m\033[22m\033[39m'


def test_piping_constitency():
    """
    Test piping a string into a style does the same as the method call
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)

    string_piped = colorful.bold & colorful.red | 'No, I am your father'
    string_called = (colorful.bold & colorful.red)('No, I am your father')
    string_called_single = colorful.bold_red('No, I am your father')

    assert str(string_piped) == str(string_called) == str(string_called_single)


def test_colorfulstring_format_protocol():
    """
    Test format protocol for colorful string
    """
    colorful = core.Colorful(colormode=terminal.ANSI_8_COLORS)
    s = colorful.black('father')

    string = 'No, I am your {0}'.format(s)
    assert string == 'No, I am your \033[30mfather\033[39m\033[26m'

    string = colorful.red('No, I am your {0}')
    formatted = string.format(s)
    assert str(formatted) == '\033[31mNo, I am your \033[30mfather\033[39m\033[26m\033[39m'



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import ast
import os
import sys
import codecs
from setuptools import setup, find_packages

# These python versions are explicitly not supported
# by colorful. This is mostly because of the incompatiblities
# with unicode strings. If there is an urgent reason why
# to support it after all or if you have a quick fix
# please open an issue on GitHub.
EXPL_NOT_SUPPORTED_VERSIONS = ((3, 0), (3, 1), (3, 2), (3, 3))

if sys.version_info[0:2] in EXPL_NOT_SUPPORTED_VERSIONS:
    raise SystemExit("colorful does explicitly not support the following python versions "
                     "due to big incompatibilities: {0}".format(EXPL_NOT_SUPPORTED_VERSIONS))


#: Holds the root dir for the project.
PROJECT_ROOT = os.path.dirname(__file__)


class VersionFinder(ast.NodeVisitor):

    def __init__(self):
        self.version = None

    def visit_Assign(self, node):
        try:
            if node.targets[0].id == '__version__':
                self.version = node.value.s
        except:
            pass


def read_version():
    """Read version from __init__.py without loading any files"""
    finder = VersionFinder()
    path = os.path.join(PROJECT_ROOT, 'colorful', '__init__.py')
    with codecs.open(path, 'r', encoding='utf-8') as fp:
        file_data = fp.read().encode('utf-8')
        finder.visit(ast.parse(file_data))

    return finder.version


with open("README.md", "r") as desc_file:
    long_description = desc_file.read()


#: Holds the requirements for colorful
requirements = [
    'colorama;platform_system=="Windows"'
]


if __name__ == '__main__':

    setup(
        name='colorful',
        version=read_version(),
        description='Terminal string styling done right, in Python.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='http://github.com/timofurrer/colorful',
        author='Timo Furrer',
        author_email='tuxtimo@gmail.com',
        maintainer='Timo Furrer',
        maintainer_email='tuxtimo@gmail.com',
        include_package_data=True,
        package_data={'': ['colorful/data/*.txt', 'README.md']},
        packages=find_packages(exclude=['*tests*']),
        install_requires=requirements,
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy'
        ]
    )



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""
import re
import sys

PY2 = sys.version_info.major == 2

if PY2:
    UNICODE = unicode  # noqa
else:
    UNICODE = str

# Catch error in case sys.stdout was patched with an object that doesn't provide
# the 'encoding' attribute.
try:
    DEFAULT_ENCODING = sys.stdout.encoding or 'utf-8'
except AttributeError:
    DEFAULT_ENCODING = 'utf-8'


def hex_to_rgb(value):
    """
    Convert the given hex string to a
    valid RGB channel triplet.
    """
    value = value.lstrip('#')
    check_hex(value)

    length = len(value)
    step = int(length / 3)
    return tuple(int(value[i:i+step], 16) for i in range(0, length, step))


def check_hex(value):
    """
    Check if the given hex value is a valid RGB color

    It should match the format: [0-9a-fA-F]
    and be of length 3 or 6.
    """
    length = len(value)
    if length not in (3, 6):
        raise ValueError('Hex string #{} is too long'.format(value))

    regex = r'[0-9a-f]{{{length}}}'.format(length=length)
    if not re.search(regex, value, re.I):
        raise ValueError('Invalid Hex String: #{}'.format(value))



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

# This module provides the X11 rgb.txt as python dictionary.
# See https://en.wikipedia.org/wiki/X11_color_names for more information

import os

from . import utils


def parse_rgb_txt_file(path=None):
    """
    Parse the given rgb.txt file into a Python dict.

    :param str path: the path to the X11 rgb.txt file
    """
    #: Holds the generated color dict
    color_dict = {}

    # get the default X11 rgb.txt file shipped with colorful
    if path is None:
        path = os.path.join(os.path.dirname(__file__), 'data', 'rgb.txt')

    def __make_valid_color_name(name):
        """
        Convert the given name into a valid colorname
        """
        if len(name) == 1:
            name = name[0]
            return name[:1].lower() + name[1:]

        return name[0].lower() + ''.join(word.capitalize() for word in name[1:])

    with open(path, 'r') as rgb_txt:
        for line in rgb_txt:
            line = line.strip()
            if not line or line.startswith('!'):
                continue  # skip comments

            parts = line.split()
            color_dict[__make_valid_color_name(parts[3:])] = (int(parts[0]),
                                                              int(parts[1]),
                                                              int(parts[2]))

    return color_dict


def sanitize_color_palette(colorpalette):
    """
    Sanitze the given color palette so it can
    be safely used by Colorful.

    It will convert colors specified in hex RGB to
    a RGB channel triplet.
    """
    new_palette = {}
    for key, value in colorpalette.items():
        if isinstance(value, str):
            # we assume it's a hex RGB value
            value = utils.hex_to_rgb(value)
        new_palette[key] = value

    return new_palette



# -*- coding: utf-8 -*-

"""
    colorful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

from __future__ import unicode_literals

import sys


#: Holds a flag whether Python 2 is used or not
PY2 = sys.version_info.major == 2

if PY2:
    string_types = (str, unicode)
else:
    string_types = (str,)
